#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import hashlib
import os

from .metadata import DatasetMetaData, ExperimentMetaData
from .singleton import SingletonDecorator
from .theorem import Theorem


@SingletonDecorator
class DatasetTracker:

    def __init__(self) -> None:
        super().__init__()
        self.datasets = list()

    def dataset_hash(self, cwd: str):
        full_hash_str = ""

        for ds in self.datasets:
            dsm, _ = DatasetMetaData().get_by_ds_name(ds, cwd=cwd)
            full_hash_str += dsm.file_hash

        full_hash = hashlib.sha1()
        full_hash.update(full_hash_str.encode())
        return full_hash.hexdigest()


def use_datasets(*args, cwd: str = '', ):
    if cwd == '':
        cwd = os.getcwd()

    dt = DatasetTracker()

    for ds in args:
        if not check_ds_exists(ds):
            raise Exception(
                f"Dataset '{ds}' not found. Please make sure it has been added via 'au data add dataset_name'"
            )

    for ds in args:
        dt.datasets.append(ds)

    is_new, str_hash = is_new_dataset(cwd)

    if is_new:
        Theorem().dataset_did_change(str_hash)


def check_ds_exists(file_name: str, cwd: str = '') -> bool:
    try:
        dmd = DatasetMetaData()
        if cwd != '':
            dmd.cwd = cwd
        dmd.get_by_ds_name(file_name=file_name, cwd=cwd)
        return True
    except FileNotFoundError:
        return False


def is_new_dataset(cwd: str) -> (bool, str):
    last_experiment_run = ExperimentMetaData().get_latest()
    currently_using = DatasetTracker().dataset_hash(cwd)

    if last_experiment_run is None:
        return True, currently_using
    elif last_experiment_run and last_experiment_run.dataset_hash != currently_using:
        return True, currently_using

    return False, ""



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##
import logging

from .metadata import CodeMetaData, generate_src_files_hash


def is_new_code() -> (bool, str):
    # Generate the current code hash string
    current_code_hash = generate_src_files_hash()

    # Get the latest saved version of the code
    latest = CodeMetaData().get_latest()

    def save_code_metadata(parent_hash: str = None) -> None:
        cmd = CodeMetaData()
        cmd.file_path_and_hash = current_code_hash
        cmd.parent_hash = parent_hash
        if parent_hash:
            logging.debug(f"Saving CodeMetaData latest file_path_and_hash: {latest.file_path_and_hash} "
                          f"and with file_path_and_hash: {current_code_hash} ")
        else:
            logging.debug(f"Saving CodeMetaData without latest and with file_path_and_hash: {current_code_hash} ")
        cmd.save()

    # If we don't have a latest. then this is the first run.
    if not latest:
        save_code_metadata()
        return True, current_code_hash

    # If we do have a latest and the file hash has changed.
    if latest and latest.file_path_and_hash != current_code_hash:
        save_code_metadata(latest.file_path_and_hash)
        return True, current_code_hash

    # Code has not changed and is the same as it was in latest
    return False, latest.file_path_and_hash



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

import argparse
import logging
import ntpath
import os
import shutil
import sys
import zipfile
from pathlib import Path

from . import constants as cons, base, git
from .dataset_tracker import check_ds_exists
from .env_builder import create_temporary_env, install_packages
from .metadata import DatasetMetaData, MetricsMetaData, ExperimentMetaData, RequirementsMetaData
from .utils import make_safe_filename, is_unnitest_running, dic_to_str, copy_dir_and_files


def run_init() -> None:
    if os.path.exists(base.DEFAULT_DIRS[0] / cons.INITIAL_COMMIT_FILE):
        logging.error("Aurum was previously initialized. Aborting.")
        sys.exit(1)

    logging.info("Initializing git...")
    git.init()

    logging.info("Creating .gitignore...")
    create_gitignore()

    logging.info("Initializing aurum...")
    au_init()

    logging.debug(f"Repository {base.get_cwd()} initialized.")


def run_add(parsed_result: argparse.Namespace, selected_dir: str) -> int:
    files_added = 0

    for f in parsed_result.files:

        # Check if file is a network address so we download it.
        # if re.search("https?://(?:[-\w.]|(?:%[\da-fA-F]{2}))+", f):
        #     filename = ntpath.basename(f)
        #
        #     if check_ds_exists(filename, cwd=selected_dir):
        #         logging.warning(f"Dataset {filename} already added to the system. Skipping...")
        #         continue
        #
        #     logging.info(f"Downloading {filename} from {f}")
        #     _, file_extension = os.path.splitext(f)
        #     local_path = f"{make_safe_filename(filename)}{file_extension}"
        #     download_file_from_web(f, local_path)
        #     logging.info(f"Downloaded {filename} to {local_path}")
        #     f = local_path
        # else:
        if check_ds_exists(f, cwd=selected_dir):
            logging.warning(f"Dataset {f} already added to the system. Skipping...")
            continue

        full_f = os.path.join(selected_dir, f)
        logging.debug(f"full f is {full_f}")

        if Path(git.get_git_repo_root(selected_dir)) != Path(selected_dir):
            check_file(f, cwd=git.get_git_repo_root(selected_dir))
            # Use relative from here on out.
            f = full_f.split(git.get_git_repo_root(selected_dir))[1]
        else:
            check_file(f, cwd=selected_dir)

        mdf = DatasetMetaData()
        mdf.file_name = f
        mdf.size = os.path.getsize(full_f)
        meta_data_file_name = mdf.save(cwd=git.get_git_repo_root(selected_dir))
        git.add(meta_data_file_name, cwd=git.get_git_repo_root(selected_dir))
        files_added += 1

    if files_added > 0:
        git.commit(f"Aurum added the metadata files for dataset(s) {', '.join(parsed_result.files)} to the project.",
                   cwd=selected_dir)
        print(f"\nAurum added the metadata files for the following datasets: {', '.join(parsed_result.files)}\n")
        print(f"-" * 80)
        print(f"If you'd like to add the actual dataset files to the project's repository,\n"
              f"you'll need to run the following commands:\n")

        for dataset_filename in parsed_result.files:
            print(f"git add {dataset_filename}")

        print(f"git commit -a -m 'Adding dataset files to project'\n")

    return files_added


def run_rm(parsed_result: argparse.Namespace, selected_dir: str) -> None:
    for filepath in parsed_result.files:

        filepath = check_file(filepath, cwd=git.get_git_repo_root(selected_dir))

        logging.info(f"Removing {filepath} from git")
        git.rm(filepath, soft_delete=parsed_result.soft_delete)
        logging.info(f"{filepath} removed from git")

        _, meta_data_path = DatasetMetaData().get_by_ds_name(cwd=git.get_git_repo_root(selected_dir),
                                                             file_name=filepath)

        logging.info(f"Removing meta data '{meta_data_path}' and removing from git.")

        git.rm(meta_data_path, soft_delete=parsed_result.soft_delete)

        # might have been removed by git, might not.
        if os.path.exists(meta_data_path):
            os.remove(meta_data_path)

        # remove parent dir if empty to avoid lots of empty dirs.
        parent_dir = os.path.join(selected_dir, cons.REPOSITORY_DIR, cons.DATASET_METADATA_DIR,
                                  make_safe_filename(filepath))
        if len(os.listdir(parent_dir)) <= 1:
            shutil.rmtree(parent_dir, ignore_errors=True)

        logging.info(f"Removed meta data '{meta_data_path}' and removed from git.")


def run_load(parsed_result: argparse.Namespace, skip_package_install: bool = False) -> None:
    logging.info(f"Attempting to load experiment with tag: {parsed_result.tag}")

    experiment_dir = os.path.join(cons.REPOSITORY_DIR, cons.EXPERIMENTS_METADATA_DIR)

    experiments = os.listdir(experiment_dir)
    if f"{parsed_result.tag}.json" not in experiments:
        msg = f"Unknown experiment tag: {parsed_result.tag} "
        logging.debug(msg)
        raise Exception(msg)

    emd = ExperimentMetaData(os.path.join(experiment_dir, f"{parsed_result.tag}.json"))

    requirements_metadata_dir = os.path.join(cons.REPOSITORY_DIR, cons.REQUIREMENTS_METADATA_DIR)

    for r in os.listdir(requirements_metadata_dir):
        if r == cons.KEEP_FILE:
            continue

        rmd = RequirementsMetaData(os.path.join(requirements_metadata_dir, r))
        if emd.requirements_hash == rmd.file_hash:
            git.run_git("checkout", "-B", f"{parsed_result.tag}")

            virtual_env_dir = create_temporary_env(parsed_result.tag)

            contents = list()

            for line in rmd.contents.splitlines():
                # Skip ourselves
                if "aurum" in line:
                    continue

                contents.append(line)

            if not skip_package_install:
                install_packages(virtual_env_dir, contents)

            print(f"Please activate your new virtual environment at: {virtual_env_dir}")

            if sys.platform == 'win32':
                bin_name = 'Scripts'
            else:
                bin_name = 'bin'

            full_virtual_env = os.path.join(virtual_env_dir, bin_name, "activate")

            print(f"Run 'source {full_virtual_env}'")

            break


def create_default_dirs() -> None:
    for path in base.DEFAULT_DIRS:
        if path.exists() and path.parts[-1] != cons.REPOSITORY_DIR and os.listdir(path) != ['.keep']:
            logging.error(f"Can't create {path} directory. Already exists.")
            sys.exit(1)
        logging.debug(f"Creating dir {path}")

        os.makedirs(path, exist_ok=True)
        Path(path, '.keep').touch()  # Needed to allow adding an empty directory to git


def create_gitignore() -> None:
    gitignore_path = base.DEFAULT_DIRS[0].parent / cons.GITIGNORE_FILE
    with open(os.path.join(os.path.dirname(__file__), cons.GITIGNORE_TEMPLATE_FILE), 'rb') as template_file:
        template = template_file.read()
        with open(gitignore_path, 'ab') as gitignore:
            gitignore.write(b"\n" + template)
    git.add(gitignore_path)


def au_init() -> None:
    create_default_dirs()
    git.add_dirs(base.DEFAULT_DIRS)
    logging.info("Adding directories to git...")

    # TODO: Move this code to the test case.
    # This is not a good idea.
    if not is_unnitest_running():
        stdout, stderr = git.commit('Initial Commit')
        logging.info("Initial commit")

        initial_commit = base.DEFAULT_DIRS[0] / cons.INITIAL_COMMIT_FILE
        open(initial_commit, 'wb').write(stdout.split()[2][:-1])
        git.add(initial_commit)
        git.commit("Recording initial commit file.")


def check_file(file_path: str, cwd: str = '') -> str:
    """
    Checks if path exists, is a file, and if absolute if can be made into a au relative path.
    If not raises SystemExit.
    """

    if cwd == '':
        cwd = git.get_git_repo_root()

    full_path = os.path.join(cwd, file_path)

    if not os.path.exists(full_path) and not os.path.exists(file_path):
        logging.error(f"Path '{file_path}' does not exist or is not in the repository: {full_path}")
        sys.exit(1)

    # if not os.path.isfile(full_path):
    #     logging.error(f"Path '{file_path}' must be a file")
    #     sys.exit(1)
    #
    # if os.path.isabs(file_path):
    #
    #     if str(base.get_cwd()) in file_path:
    #         file_path = file_path.split(str(base.get_cwd()), 1)[1][1:]
    #     else:
    #         logging.error(f"File '{file_path}' is not relative to au repository")
    #         sys.exit(1)

    return file_path


def display_metrics(experiment_ids: list) -> None:
    metrics_path = os.path.join(cons.REPOSITORY_DIR, cons.METRICS_METADATA_DIR)
    for path in os.listdir(metrics_path):
        if cons.KEEP_FILE not in path:
            metrics_metadata = MetricsMetaData(os.path.join(metrics_path, path))
            if len(experiment_ids) > 0:
                if metrics_metadata.experiment_id in experiment_ids:
                    print(dic_to_str(metrics_metadata.metrics, f'Experiment id: {metrics_metadata.experiment_id}'))
            else:
                print(dic_to_str(metrics_metadata.metrics, f'Experiment id: {metrics_metadata.experiment_id}'))


def export_experiment(parsed_args: argparse.Namespace) -> None:
    remove_dirs = []
    dataset_path = None
    repo_dir = cons.REPOSITORY_DIR
    dataset_metadata = DatasetMetaData().get_latest()
    root_path = git.get_git_repo_root()

    if dataset_metadata:

        if parsed_args.no_data:
            remove_dirs.append(os.path.join(repo_dir, dataset_metadata.file_name))
        else:
            dataset_path = dataset_metadata.file_name

    if parsed_args.no_logs:
        remove_dirs.append(os.path.join(root_path, cons.LOGS_DIR))
    if parsed_args.no_metrics:
        remove_dirs.append(os.path.join(repo_dir, cons.METRICS_METADATA_DIR))

    destiny_path = os.path.join(parsed_args.tag)
    if not Path(destiny_path).exists():
        os.makedirs(destiny_path)

    for path in base.DEFAULT_DIRS:
        if path.as_posix() in remove_dirs:
            continue
        copy_dir_and_files(path, os.path.join(destiny_path, ntpath.basename(path)))

    if dataset_path:
        copy_dir_and_files(dataset_path, os.path.join(destiny_path, ntpath.basename(dataset_path)))

    filename = os.path.join(root_path, parsed_args.tag + '.zip')
    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zip_ref:
        for root, dirs, files in os.walk(destiny_path):
            for file in files:
                zip_ref.write(os.path.join(root, file))

    shutil.rmtree(destiny_path, ignore_errors=True)



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


from .singleton import SingletonDecorator
import uuid


@SingletonDecorator
class Theorem:
    """
    Singleton class responsible act as a single source of truth regarding the experiment change's state
    Eg. inform that the source-code has changed
    """

    def __init__(self):
        self.requirements_changed = False
        self.code_changed = False
        self.dataset_changed = False
        self.parameters_changed = False
        self.experiment_id = str(uuid.uuid4())

    def has_any_change(self):
        return self.requirements_changed is not False or self.code_changed is not False or \
               self.dataset_changed is not False or self.parameters_changed is not False

    def requirements_did_change(self, requirements_metadata_location_hash: str) -> None:
        self.requirements_changed = requirements_metadata_location_hash

    def code_did_change(self, code_metadata_location_hash: str) -> None:
        self.code_changed = code_metadata_location_hash

    def dataset_did_change(self, dataset_metadata_location_hash: str) -> None:
        self.dataset_changed = dataset_metadata_location_hash

    def parameters_did_change(self, parameters_metadata_location_hash: str) -> None:
        self.parameters_changed = parameters_metadata_location_hash



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

import argparse
import json
import os

from . import constants as cons
from .metadata import ParameterMetaData
from .singleton import SingletonDecorator


@SingletonDecorator
class ExperimentArgParser:
    def __init__(self):

        # TODO: Change this description so that it reflects what the ExperimentArgParser really is.
        description = """Aurum is a new and simplified approach for data scientists to
        keep track of data and code without having to get another PhD for it. Aurum
        keeps track of all code and data changes, and lets you easily reproduce any
        experiment as well as easily compare metrics across experiments.
        """
        epilog = "And that's how you make your live easier. You're welcome."

        self.known_params = None
        self.unknown_params = None

        self.parser = argparse.ArgumentParser(description=description, epilog=epilog, add_help=True)

        self.parser.add_argument("--verbose", help="increase output verbosity", action="store_true")
        self.parser.add_argument("--no-tracking", help="Don't store any metadata about this experiment",
                                 action="store_true")

        try:
            known_params = load_parameters()
        except FileNotFoundError:
            known_params = {}

        for param in known_params.keys():
            self.parser.add_argument(f'-{param}', required=False, default=known_params[param])

        self.parse_args()

        # TODO: Save preference on verbose or dry run into the instance for easy access

    def parse_args(self):
        self.known_params, self.unknown_params = self.parser.parse_known_args()


def load_parameters() -> dict:
    metadata = ParameterMetaData().get_latest()

    if metadata and metadata.file_name:
        filepath = os.path.join(cons.REPOSITORY_DIR, cons.PARAMETER_METADATA_DIR, metadata.file_name)
    else:
        filepath = None

    if filepath is not None:
        with open(filepath, 'r') as f:
            root_json = json.loads(f.read())
            return json.loads(root_json['parameters'])
    else:
        return {}



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import os

from .metadata import MetaData
from .. import constants as cons
from .. import git
from ..utils import gen_dict_hash


class ExperimentMetaData(MetaData):

    def __init__(self, file_name: str = '') -> None:
        self.name = None
        self.dataset_hash = None
        self.code_hash = None
        self.parameter_hash = None
        self.requirements_hash = None
        self.logging_hash = None
        self.weights_hash = None
        super().__init__(file_name)

    def save(self, destination: str) -> str:
        self.file_hash = self.obj_hash()

        parent = self.get_latest()

        if parent and parent.file_hash != self.file_hash:
            self.parent_hash = parent.file_hash

        return super().save(destination)

    def get_dir(self):
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.EXPERIMENTS_METADATA_DIR,
        )

    def obj_hash(self):
        return gen_dict_hash(
            {
                'code_hash': self.code_hash,
                'dataset_hash': self.dataset_hash,
                'parameter_hash': self.parameter_hash,
                'parent_hash': self.parent_hash,
                'requirements_hash': self.requirements_hash,
            }
        )



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import hashlib
import logging
import os
import subprocess
from . import constants as cons

from .metadata import RequirementsMetaData


def is_new_requirements(cwd: str = '') -> (bool, str):
    """
    Run a pip freeze and create a hash to be saved in the requirements metadata, remember that we will also need to
    record the parent requirements (latest by date, if it exists) as well as all the contents of the pip freeze list
    in a alphabetically sorted list.

    Make sure that the metadata is a linked list and can be easily browsed to find the parent even if you don't know
    who the parent is (similar to all other metadata implemented in the system).

    If there is a change in the results from freeze, this module should indicate that this is a new experiment, and
    create the metadata to document this. If there is no previous metadata (experiment is being run for the first time)
    then the code should proceed as if this is a brand new experiment, except that parent will be None.
    """

    if cwd == '':
        cwd = os.getcwd()

    process = subprocess.run(
        "pip freeze",
        shell=True,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=cwd,
    )
    output = process.stdout
    logging.debug(f"\nInstalled packages: \n {output} \n\n")
    packages_hash = hashlib.sha1()
    packages_hash.update(output)
    packages_hash = packages_hash.hexdigest()

    latest_mdf = RequirementsMetaData().get_latest(
        subdir_path=os.path.join(cwd, cons.REPOSITORY_DIR, cons.REQUIREMENTS_METADATA_DIR)
    )

    if not latest_mdf:
        logging.debug("This is a new requirements")
        rmd = RequirementsMetaData()
        rmd.file_hash = packages_hash
        rmd.contents = output.decode()
        rmd.save(cwd=cwd)
        return True, rmd.file_hash

    elif latest_mdf and latest_mdf.file_hash != packages_hash:
        logging.debug("This are changed requirements")
        rmd = RequirementsMetaData()
        rmd.file_hash = packages_hash
        rmd.parent_hash = latest_mdf.file_hash
        rmd.contents = output.decode()
        rmd.save(cwd=cwd)
        return True, rmd.file_hash

    return False, ""



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import json
import logging
import os

from .metadata import MetaData, gen_meta_file_name_from_hash
from .. import constants as cons, git
from ..utils import gen_file_hash, make_safe_filename


class DatasetMetaData(MetaData):

    def __init__(self, file_name: str = '') -> None:
        super().__init__(file_name)
        self.size = 0

    def save(self, cwd: str, destination: str = None, ) -> str:
        meta_data_path = os.path.join(cons.REPOSITORY_DIR, cons.DATASET_METADATA_DIR)
        destination = gen_meta_file_name_from_hash(
            meta_data_str=json.dumps(self.serialize()),
            file_name=self.file_name,
            path=meta_data_path,
        )

        self.file_hash = gen_file_hash(os.path.join(cwd, self.file_name))

        old_dataset_metadata = DatasetMetaData().get_latest(
            subdir_path=os.path.join(cwd, '.au', cons.DATASET_METADATA_DIR))

        if old_dataset_metadata and self.file_hash != old_dataset_metadata.file_hash:
            self.parent_hash = old_dataset_metadata.file_hash

        return super().save(destination, cwd)

    def get_dir(self):
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.DATASET_METADATA_DIR,
        )

    def get_by_ds_name(self, file_name: str, cwd: str):

        meta_data_dir = os.path.join(cwd, self.get_dir(), make_safe_filename(file_name))

        for mdf in os.listdir(meta_data_dir):

            # Ignore keep files.
            if cons.KEEP_FILE in mdf:
                continue

            meta_data_path = os.path.join(meta_data_dir, mdf)
            mdo = DatasetMetaData(meta_data_path)

            if mdo.file_name == file_name:
                return mdo, meta_data_path

        raise FileNotFoundError(f"Metadata not found for {file_name}")



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

__author__ = "Adriano Marques, Nathan Martins, Thales Ribeiro"
__copyright__ = "Copyright (C) 2019 Exponential Ventures LLC"
__license__ = "GNU LESSER GENERAL PUBLIC LICENSE 2.0"
__url__ = "https://github.com/exponential-ventures/aurum"
__version__ = "0.1"

import logging
import os
import sys
from pathlib import PurePosixPath, PureWindowsPath

from .au import main
from .base import execute_commands, parameters, register_metrics, save_metrics, end_experiment, save_weights, load_weights
from .code_tracker import is_new_code
from .dataset_tracker import use_datasets
from .dry_run import Dehydrator
from .experiment_parser import ExperimentArgParser
from .logging_tracker import LoggingTracker
from .package_tracker import is_new_requirements
from aurum.theorem import Theorem
from .time_tracker import time_tracker
from .utils import check_inside_au

# TODO: Check if this works on windows.
if os.name == 'nt':
    command = PureWindowsPath(sys.argv[0]).name
else:
    command = PurePosixPath(sys.argv[0]).name

if command is not 'au' and 'unittest' not in command:

    check_inside_au()
    parser = ExperimentArgParser()

    if parser.known_params.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    if parser.known_params.no_tracking is True:
        Dehydrator().on()

    LoggingTracker()

    requirements_changed, r_hash = is_new_requirements()
    if requirements_changed:
        Theorem().requirements_did_change(r_hash)

    code_changed, c_hash = is_new_code()
    if code_changed:
        Theorem().code_did_change(c_hash)

__all__ = [
    execute_commands,
    parameters,
    register_metrics,
    use_datasets,
    end_experiment,
    save_weights,
    load_weights,
]



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

import logging
import os
import subprocess
import sys


class GitCommandError(Exception):
    pass


def check_git():
    process = run_git("--version")
    output = process.communicate()[0].decode("utf-8")

    if "git version" not in output:
        logging.error("Git not installed. Aborting.")
        sys.exit(1)

    logging.debug("Git found.")


def running_from_git_repo(cwd: str = "") -> bool:
    if cwd == "":
        cwd = os.getcwd()

    process = subprocess.Popen(
        ["git", "rev-parse", "--is-inside-work-tree"],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        cwd=cwd,
    )
    output, _ = process.communicate()
    output = output.decode("utf-8")

    return str(output) == "true\n"


def get_git_repo_root(cwd: str = '') -> str:
    if cwd == "":
        cwd = os.getcwd()

    if not running_from_git_repo(cwd):
        logging.debug("Not running from a git repo")
        return ""

    process = run_git(
        "rev-parse",
        "--show-toplevel",
        cwd=cwd
    )

    output, _ = process.communicate()
    return output.decode("utf-8").replace("\n", "/")


def init():
    p = run_git("init")
    _, err = p.communicate()
    if p.returncode != 0:
        raise GitCommandError(f"Failed to run 'git init': {err}")


def rm(filepath, soft_delete: bool = True):
    if soft_delete:
        run_git("rm", "--cached", filepath).wait()
    else:
        run_git("rm", filepath).wait()


def add_dirs(dirs: list) -> None:
    for path in dirs:
        process = run_git('add', path)
        _, error = process.communicate()

        if process.returncode != 0:
            raise GitCommandError(f"Failed to run 'git add {path}': {error}")


def tag(experiment_id: str, message: str) -> None:
    process = run_git('tag', '-a', experiment_id, '-m', message)
    _, error = process.communicate()

    if process.returncode != 0:
        raise GitCommandError(f"Failed to run 'git tag -a {experiment_id} -m {message}': {error}")


def commit(commit_message: str, secondary_msg: str = '', cwd: str = '') -> (str, str):
    if secondary_msg is not '':
        process = run_git('commit', '-am', commit_message, '-m', secondary_msg, cwd=cwd)
    else:
        process = run_git('commit', '-am', commit_message, cwd=cwd)

    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise GitCommandError(
            f"Failed to run 'git commit -am {commit_message} -m {secondary_msg}': {stdout} -- {stderr}")

    return stdout, stderr


def last_commit_hash() -> str:
    process = run_git('rev-parse', 'HEAD')

    output, error = process.communicate()

    if process.returncode != 0:
        raise GitCommandError(f"Failed to run 'git rev-parse': {error}")

    return output.decode('utf-8').replace('\n', '')


def current_branch_name() -> str:
    process = run_git('rev-parse', '--abbrev-ref', 'HEAD')
    output, error = process.communicate()

    if process.returncode != 0:
        raise GitCommandError(f"Failed to run 'git rev-parse --abbrev-ref HEAD': {error}")

    return output.decode('utf-8').replace('\n', '')


def push() -> str:
    if not has_remote():
        return ''
    sub = run_git('push', '--all')
    output, error = sub.communicate()
    if sub.returncode != 0:
        raise GitCommandError(f"Failed to run 'git push --all': {error}")

    return output.decode('utf-8').replace('\n', '')


def push_tags() -> str:
    if not has_remote():
        return ''
    sub = run_git('push', '--tags')
    output, error = sub.communicate()
    if sub.returncode != 0:
        raise GitCommandError(f"Failed to run 'git push --tags': {error}")
    return output.decode('utf-8').replace('\n', '')


def has_remote() -> bool:
    sub = run_git('remote', 'show')
    output, error = sub.communicate()
    if sub.returncode != 0:
        raise GitCommandError(f"Failed to run 'git remote show': {error} /n Please configure a remote repository")

    output = output.decode('utf-8').replace('\n', '')
    return bool(output and output.strip())


def add(*filenames: str, cwd: str = '') -> str:
    sub = run_git('add', *filenames, cwd=cwd)
    stdout, stderr = sub.communicate()

    if sub.returncode != 0:
        raise GitCommandError(f"Failed to run 'git add {filenames}': {stderr}")

    return stdout.decode('utf-8').replace('\n', '')


def run_git(*args, cwd: str = ''):
    if cwd == '':
        cwd = os.getcwd()
    return subprocess.Popen(["git"] + list(args), stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import logging
import os
import os.path
import subprocess
import sys
import tempfile
import venv
from subprocess import Popen, PIPE
from threading import Thread
from urllib.parse import urlparse
from urllib.request import urlretrieve


class ExtendedEnvBuilder(venv.EnvBuilder):
    """
    This builder installs setuptools and pip so that you can pip or
    easy_install other packages into the created virtual environment.

    :param progress: If setuptools or pip are installed, the progress of the
                     installation can be monitored by passing a progress
                     callable. If specified, it is called with two
                     arguments: a string indicating some progress, and a
                     context indicating where the string is coming from.
                     The context argument can have one of three values:
                     'main', indicating that it is called from virtualize()
                     itself, and 'stdout' and 'stderr', which are obtained
                     by reading lines from the output streams of a subprocess
                     which is used to install the app.

                     If a callable is not specified, default progress
                     information is output to sys.stderr.
    """

    def post_setup(self, context):
        """
        Set up any packages which need to be pre-installed into the
        virtual environment being created.

        :param context: The information for the virtual environment
                        creation request being processed.
        """
        os.environ['VIRTUAL_ENV'] = context.env_dir

        self.install_setuptools(context)

        # Can't install pip without setuptools
        self.install_pip(context)

    def reader(self, stream, context):
        """
        Read lines from a subprocess' output stream and either pass to a progress
        callable (if specified) or write progress information to sys.stderr.
        """

        while True:
            s = stream.readline()
            if not s:
                break

            sys.stdout.write(s.decode('utf-8'))
            sys.stdout.flush()
        stream.close()

    def install_script(self, context, name, url):
        _, _, path, _, _, _ = urlparse(url)
        fn = os.path.split(path)[-1]
        binpath = context.bin_path
        distpath = os.path.join(binpath, fn)
        # Download script into the virtual environment's binaries folder
        urlretrieve(url, distpath)

        term = '\n'

        sys.stdout.write(f'Installing {name}| ...%{term}')

        # Install in the virtual environment
        args = [context.env_exe, fn]
        p = Popen(args, stdout=PIPE, stderr=PIPE, cwd=binpath)
        t1 = Thread(target=self.reader, args=(p.stdout, 'stdout'))
        t1.start()
        t2 = Thread(target=self.reader, args=(p.stderr, 'stderr'))
        t2.start()
        p.wait()
        t1.join()
        t2.join()

        sys.stdout.write('done.\n')
        # Clean up - no longer needed
        os.unlink(distpath)

    def install_setuptools(self, context):
        """
        Install setuptools in the virtual environment.

        :param context: The information for the virtual environment
                        creation request being processed.
        """
        url = 'https://bitbucket.org/pypa/setuptools/downloads/ez_setup.py'
        self.install_script(context, 'setuptools', url)
        # clear up the setuptools archive which gets downloaded
        pred = lambda o: o.startswith('setuptools-') and o.endswith('.tar.gz')
        files = filter(pred, os.listdir(context.bin_path))
        for f in files:
            f = os.path.join(context.bin_path, f)
            os.unlink(f)

    def install_pip(self, context):
        """
        Install pip in the virtual environment.

        :param context: The information for the virtual environment
                        creation request being processed.
        """
        url = 'https://raw.github.com/pypa/pip/master/contrib/get-pip.py'
        self.install_script(context, 'pip', url)


def create_temporary_env(name: str):
    compatible = True
    if sys.version_info < (3, 3):
        compatible = False
    elif not hasattr(sys, 'base_prefix'):
        compatible = False
    if not compatible:
        raise ValueError('This script is only for use with '
                         'Python 3.3 or later')

    if os.name == 'nt':
        use_symlinks = False
    else:
        use_symlinks = True

    dir_name = os.path.join(tempfile.gettempdir(), name)

    clear = False

    if os.path.exists(dir_name):
        clear = True
    else:
        os.makedirs(dir_name)

    builder = venv.EnvBuilder(symlinks=use_symlinks, clear=clear, with_pip=True)
    builder.create(dir_name)

    return dir_name


def install_packages(env_dir, pip_file_contents: list):
    python_location = get_py_exe(env_dir)

    logging.debug(f"Running install_packages with python_location: {python_location}")

    update_pip(env_dir)

    logging.debug(f"Installing: {pip_file_contents}")

    proc = subprocess.Popen(
        [python_location, "-m", "pip", "install", *pip_file_contents],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    result, err = proc.communicate()

    if proc.returncode != 0:
        raise Exception(f"Failed to run pip install({proc.returncode}): {err}")

    return result.decode()


def get_py_exe(env_dir):
    executable = getattr(sys, '_base_executable', sys.executable)
    dirname, exe_name = os.path.split(os.path.abspath(executable))

    if sys.platform == 'win32':
        bin_name = 'Scripts'
    else:
        bin_name = 'bin'

    bin_path = os.path.join(env_dir, bin_name)

    return os.path.join(bin_path, exe_name)


def update_pip(env_dir):
    python_location = get_py_exe(env_dir)

    logging.debug(f"Running update_pip with python_location: {python_location}")

    proc = subprocess.Popen(
        [python_location, "-m", "pip", "install", "--upgrade", "pio"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    result, err = proc.communicate()

    if proc.returncode != 0:
        raise Exception(f"Failed to run pip upgrade: {err}")

    return result.decode()



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


REPOSITORY_DIR = ".au"
KEEP_FILE = ".keep"
SOURCE_CODE_DIR = "src"
LOGS_DIR = "logs"
INITIAL_COMMIT_FILE = "initial_commit"
GITIGNORE_FILE = ".gitignore"
GITIGNORE_TEMPLATE_FILE = "gitignore_template.txt"

DATASET_METADATA_DIR = "datasets"
PARAMETER_METADATA_DIR = "parameters"
METRICS_METADATA_DIR = "metrics"
WEIGHTS_METADATA_DIR = "weights"
WEIGHTS_BINARIES_DIR = "binaries"
REQUIREMENTS_METADATA_DIR = "requirements"
CODE_METADATA_DIR = "code"
EXPERIMENTS_METADATA_DIR = "experiments"

# COMMANDS
METRICS = 'metrics'
INIT = 'init'
LOAD = 'load'
DATA = 'data'
DATA_ADD = 'add'
DATA_RM = 'rm'
EXPERIMENT_IDS = 'experiment_ids'
EXPORT_TAG = 'export'



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

import hashlib
import json
import logging
import os
import sys
import time
import shutil
import errno
import urllib.request
from collections import OrderedDict
from pathlib import PurePosixPath, PureWindowsPath

from . import constants as cons
from . import git


def gen_file_hash(file_name):
    sha1 = hashlib.sha1()
    buf_size = 65536  # lets read stuff in 64kb chunks!

    with open(file_name, 'rb') as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()


def gen_dict_hash(dictionary: dict) -> str:
    if not isinstance(dictionary, dict):
        raise TypeError('Parameter provided is not a dictionary')

    sha1 = hashlib.sha1()
    ordered_dict = OrderedDict(dictionary.items())
    sha1.update(json.dumps(ordered_dict).encode('utf-8'))
    return sha1.hexdigest()


def make_safe_filename(s):
    def safe_char(c):
        if c.isalnum():
            return c
        else:
            return "_"

    return "".join(safe_char(c) for c in s).rstrip("_")


def size_in_gb(size):
    return f"{size / float(1 << 30)} GB"


def check_inside_au():
    path = os.path.join(git.get_git_repo_root(), cons.REPOSITORY_DIR)
    if not os.path.exists(path):
        logging.error(f"Path '.au' does not exist, please run au init \n")
        sys.exit(1)


def dic_to_str(dictionary: dict, title: str = None) -> str:
    aux_str = ""
    if title:
        aux_str += f"\n {title}"
    for k, v in sorted(dictionary.items()):
        if isinstance(v, dict):
            aux_str += dic_to_str(v, k)
        else:
            aux_str += f"\n\t {k}={v}"
    aux_str += "\n"
    return aux_str


def did_dict_change(d1, d2):
    """
    Compares two dict to detect if the first has any changes against the second.
    In case of any changes it returns True. Otherwise, it returns false.
    """
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    same = set(o for o in intersect_keys if d1[o] == d2[o])
    if (len(same) == len(d1)):
        return False

    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys

    if (len(added) > 0) or (len(removed) > 0):
        return True

    for k in intersect_keys:
        if d1[k] != d2[k]:
            return True

    return False


def is_unnitest_running() -> bool:
    if os.name == 'nt':
        command = PureWindowsPath(sys.argv[0]).name
    else:
        command = PurePosixPath(sys.argv[0]).name
    return 'unittest' in sys.modules.keys() or 'unittest' in command


def dir_files_by_last_modification_date(directory_path: str, descending_order: bool = True) -> list:
    file_last_modified_list = []

    for f in os.listdir(directory_path):
        if cons.KEEP_FILE not in f:
            file_path = os.path.join(directory_path, f)
            stats = os.stat(file_path)
            lastmod_date = time.localtime(stats[8])
            file_last_modified_list.append((lastmod_date, file_path))

    file_last_modified_list.sort()

    if descending_order:
        file_last_modified_list.reverse()  # The first is the latest modified

    return file_last_modified_list


def copy_dir_and_files(source, destiny):
    try:
        shutil.copytree(source, destiny)
    except OSError as e:
        if e.errno == errno.ENOTDIR:
            shutil.copy(source, destiny)
        else:
            logging.error(f'Directory not copied, error: {e}')


def download_file_from_web(url: str, output: str) -> None:
    with urllib.request.urlopen(url) as response, open(output, 'wb') as f:
        shutil.copyfileobj(response, f)



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import functools
import logging

from .singleton import SingletonDecorator


def log_call(func, *args, **kw):
    """
        Produce a info message which logs the *func* call.
    """
    # concatenate args and kw args transforming string values
    # from 'value' to '"value"' in order to pretty display
    func_args = []

    # concatenate positional args
    args = list(args)
    if args:
        for i, arg in enumerate(args):
            if type(arg) is str:
                args[i] = f'"{arg}"'
        func_args.extend([str(arg) for arg in args])

    # concatenate non positional args
    if kw:
        for key, value in kw.items():
            if type(value) is str:
                kw[key] = f'"{value}"'

        func_args.extend([f"{k}={v}" for k, v in kw.items()])

    logging.debug(f"[DRYRUN] call to '{func.__name__}({', '.join(func_args)})'")


def is_dry_run() -> bool:
    return Dehydrator().status


def dehydratable(func):
    """
    Decorator which makes a log the call of the target
    function without executing it.

    Example:
        >>> @dehydratable
        ... def foo(bar, baz=None):
        ...     return 42
        ...
        >>> foo('sport', baz=False)
        42
        >>> foo('sport', baz=False)
        INFO:aurum.dry_run:[DRYRUN] call to 'foo(sport, baz=False)'

    """

    @functools.wraps(func)
    def decorator(*args, **kw):
        # if dry run is disabled exec the original method
        if is_dry_run() is False:
            return func(*args, **kw)
        else:
            log_call(func, *args, **kw)
            return None

    return decorator


@SingletonDecorator
class Dehydrator:

    def __init__(self) -> None:
        self._status = False

    def on(self):
        self._status = True

    def off(self):
        self._status = False

    @property
    def status(self):
        return self._status



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##
import argparse
import logging
import platform
from pathlib import Path

import psutil
from pynvml import *

from . import constants as cons, git
from .commands import run_init, run_rm, run_add, run_load, display_metrics, export_experiment
from .metadata import (
    ParameterMetaData,
    WeightsMetaData,
    MetricsMetaData,
    ExperimentMetaData,
    DatasetMetaData,
    CodeMetaData,
    RequirementsMetaData,
)
from .theorem import Theorem
from .time_tracker import time_tracker
from .utils import size_in_gb, dic_to_str


def get_cwd():
    return Path(os.getcwd())


#######
# TODO: Figure out whether we can stick with only one of DEFAULT_DIRS or get_default_dirs.
# Not sure why we're duplicating this here. HINT: could be because we may want to have the
# paths generated during call time as opposed to right after the app is run but things are
# still changing.

DEFAULT_DIRS = [
    get_cwd() / cons.REPOSITORY_DIR,
    get_cwd() / "src",
    get_cwd() / "logs",
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.DATASET_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.REQUIREMENTS_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.PARAMETER_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.EXPERIMENTS_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.METRICS_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR, cons.WEIGHTS_BINARIES_DIR),
    get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.CODE_METADATA_DIR)
]


def get_default_dirs():
    return [
        get_cwd() / cons.REPOSITORY_DIR,
        get_cwd() / "src",
        get_cwd() / "logs",
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.DATASET_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.REQUIREMENTS_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.PARAMETER_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.EXPERIMENTS_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.METRICS_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR, cons.WEIGHTS_BINARIES_DIR),
        get_cwd() / os.path.join(cons.REPOSITORY_DIR, cons.CODE_METADATA_DIR)
    ]


def execute_commands(parser: argparse.ArgumentParser) -> None:
    parsed = parser.parse_args()

    logging.getLogger().setLevel(logging.DEBUG if parsed.verbose else logging.WARNING)

    logging.debug(f"Parser arguments: {parsed}")

    git.check_git()

    if os.getcwd().startswith(cons.REPOSITORY_DIR):
        parser.error(f"Cannot run commands from inside '.au' folder")

    if not hasattr(parsed, "subcommand"):
        parser.error(f"No command was passed in")

    if parsed.subcommand == cons.INIT:
        run_init()
    elif parsed.subcommand == cons.LOAD:
        data_command_checker(parser)
        run_load(parsed)
    elif parsed.subcommand == cons.DATA:

        if hasattr(parsed, "subcommand2") and parsed.subcommand2 == cons.DATA_RM:
            data_command_checker(parser)
            run_rm(parsed)
        if hasattr(parsed, "subcommand2") and parsed.subcommand2 == cons.DATA_ADD:
            data_command_checker(parser)
            run_add(parsed, selected_dir=os.getcwd())
        else:
            parser.error("Unknown command for data")
    elif parsed.subcommand == cons.METRICS:
        experiment_ids = []
        if parsed.experiment_ids:
            experiment_ids = parsed.experiment_ids.split(',')
        display_metrics(experiment_ids)
    elif parsed.subcommand == cons.EXPORT_TAG:
        data_command_checker(parser)
        if parsed.tag not in git.current_branch_name():
            run_load(parsed, skip_package_install=True)
        export_experiment(parsed)


def data_command_checker(parser: argparse.ArgumentParser):
    if not git.running_from_git_repo():
        parser.error(f"You are not running from inside a au repository")

    repo_root = git.get_git_repo_root()

    if not os.path.exists(os.path.join(repo_root, cons.REPOSITORY_DIR)):
        parser.error(f"Path '.au' does not exist, please run au init")


def parameters(cwd: str = '', **kwargs):
    if cwd == '':
        cwd = os.getcwd()

    from .experiment_parser import ExperimentArgParser

    p = ExperimentArgParser()

    for param, default in kwargs.items():
        if param not in p.known_params:
            p.parser.add_argument(f'-{param}', required=False, default=default)

    p.parse_args()

    if len(p.unknown_params) > 0:
        raise RuntimeError(f"Unknown parameters passed to experiment: {' '.join(p.unknown_params)}")

    new_dict = {**kwargs, **p.known_params.__dict__}

    for key in new_dict.keys():
        setattr(sys.modules['aurum'], key, new_dict[key])

    pmd = ParameterMetaData()
    pmd.parameters = new_dict

    latest_exp = ExperimentMetaData().get_latest(
        subdir_path=os.path.join(cwd, cons.REPOSITORY_DIR, cons.PARAMETER_METADATA_DIR)
    )

    if (latest_exp and latest_exp.parameter_hash != pmd.parameter_hash) or (latest_exp is None):
        Theorem().parameters_did_change(pmd.parameter_hash)
        meta_data_file_name = pmd.save(cwd=cwd)
        git_proc = git.run_git("add", meta_data_file_name)

        result = git_proc.wait()
        if result != 0:
            message = f"Unable to run 'git add {meta_data_file_name}' Exit code: {result}\n"
            if git_proc.stderr:
                message += f"{git_proc.stderr.read()}\n"
            logging.error(message)


def register_metrics(cwd: str = '', **kwargs):
    if cwd == '':
        cwd = os.getcwd()

    swap_mem = psutil.swap_memory()
    virtual_memory = psutil.virtual_memory()
    disk_usage = psutil.disk_usage('/')

    hardware_metric = {'environment': {'python_version': platform.python_version()},
                       'run time': str(time_tracker.log_time()),
                       'run time in seconds': time_tracker.log_time().total_seconds(),
                       'hardware': {
                           'swap_memory': {
                               'total': size_in_gb(swap_mem.total)
                           },
                           'virtual_memory': {
                               'total': size_in_gb(virtual_memory.total)
                           },
                           'cpu': {
                               'physical_cores': psutil.cpu_count(logical=False),
                               'total_cores': psutil.cpu_count(),
                               'frequency': psutil.cpu_freq().current,
                           },
                           'disk': {
                               'total': size_in_gb(disk_usage.total),
                           },
                           'gpu(s)': gpu_info()
                       }

                       }

    metrics = {**kwargs, **hardware_metric}
    save_metrics(cwd=cwd, **metrics)


def gpu_info():
    info = {}
    try:
        nvmlInit()
    except Exception:
        info['no-gpu'] = 'No Nvidia GPU detected'
        return info

    device_count = nvmlDeviceGetCount()

    info['driver_version'] = nvmlSystemGetDriverVersion().decode()
    info['device_count'] = device_count
    info['device'] = dict()
    for i in range(device_count):
        handle = nvmlDeviceGetHandleByIndex(i)
        memory = nvmlDeviceGetMemoryInfo(handle)

        info['device'][i] = dict()
        info['device'][i]['name'] = str(nvmlDeviceGetName(handle))

        info['device'][i]['memory'] = dict()
        info['device'][i]['memory']['total'] = str(size_in_gb(memory.total))

    nvmlShutdown()

    return info


def save_metrics(cwd: str, **kwargs):
    mmd = MetricsMetaData()
    mmd.metrics = kwargs

    meta_data_file_name = mmd.save(cwd=cwd)

    if meta_data_file_name:

        git_proc = git.run_git("add", meta_data_file_name)

        result = git_proc.wait()
        if result != 0:
            message = f"Unable to run 'git add {meta_data_file_name}' Exit code: {result}\n"
            if git_proc.stderr:
                message += f"{git_proc.stderr.read()}\n"
            logging.error(message)


def save_weights(model_encoded, cwd: str = "", ):
    meta_data_file_name = None
    wmd = WeightsMetaData()
    if cwd == "":
        cwd = os.getcwd()

    if Theorem().has_any_change():
        wmd.save_binary(model_encoded)
        meta_data_file_name = wmd.save(cwd)

    if meta_data_file_name:

        git_proc = git.run_git("add", meta_data_file_name)

        result = git_proc.wait()
        if result != 0:
            message = f"Unable to run 'git add {meta_data_file_name}' Exit code: {result}\n"
            if git_proc.stderr:
                message += f"{git_proc.stderr.read()}\n"
            logging.error(message)


def load_weights(cwd: str = ""):
    if cwd == "":
        cwd = os.getcwd()

    wmd = WeightsMetaData().get_latest(subdir_path=os.path.join(cwd, cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR))

    return wmd.load_binary()


def end_experiment() -> bool:
    commit_msg = ""

    theorem = Theorem()
    if theorem.has_any_change():

        mdt = ExperimentMetaData()

        mdt.file_name = theorem.experiment_id
        metrics_metadata = MetricsMetaData().get_latest() or MetricsMetaData()
        weights_metadata = WeightsMetaData().get_latest() or WeightsMetaData()
        parameters_metadata = ParameterMetaData().get_latest() or ParameterMetaData()
        requirements_metadata = RequirementsMetaData().get_latest() or RequirementsMetaData()
        dataset_metadata = DatasetMetaData().get_latest() or DatasetMetaData()
        code_metadata = CodeMetaData().get_latest() or CodeMetaData()
        destination = os.path.join(cons.REPOSITORY_DIR, cons.EXPERIMENTS_METADATA_DIR, f"{theorem.experiment_id}.json")

        mdt.metrics_hash = metrics_metadata.file_hash
        mdt.weights_hash = weights_metadata.file_hash
        mdt.parameter_hash = parameters_metadata.file_hash
        mdt.requirements_hash = requirements_metadata.file_hash
        mdt.code_hash = code_metadata.file_hash
        mdt.dataset_hash = dataset_metadata.file_hash

        if metrics_metadata.metrics:
            commit_msg += dic_to_str(metrics_metadata.metrics, 'Metrics')

        if parameters_metadata.parameters:
            commit_msg += dic_to_str(parameters_metadata.parameters, 'Parameters')

        if weights_metadata.file_hash:
            commit_msg += f"\n Weights hash: {weights_metadata.file_hash}"

        if requirements_metadata.file_hash:
            commit_msg += f"\n Requirements hash {requirements_metadata.file_hash}"

        if dataset_metadata.file_hash:
            commit_msg += f"\n Dataset hash: {dataset_metadata.file_hash}"
        else:
            logging.warning("No dataset detected. Please, run 'au data add' or 'au.use_dataset")

        if code_metadata.file_hash:
            commit_msg += f"\n Code hash: {code_metadata.file_hash}"
        else:
            logging.warning("Please, add the source-code under the 'src' folder")

        mdt.commit_hash = git.last_commit_hash()
        mdt.save(destination)
        git.add_dirs(get_default_dirs())
        git.commit(f"Experiment ID {theorem.experiment_id}", commit_msg)
        git.tag(theorem.experiment_id, commit_msg)

        # git push all branches
        git.push()
        # git push all tags
        git.push_tags()

        return True

    return False



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


"""
MetaData objects allow a user to:
  - dynamically add instance attributes and have them be json serializable.
  - serialize: convert json string naive(str only) MetaData object.
  - deserialize: convert MetaData object to json string.
  - save: perform a deserialization and save to file.
"""
import hashlib
import json
import logging
import os
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Collection, Mapping, Union
from uuid import UUID

from .. import constants as cons
from ..dry_run import dehydratable
from ..theorem import Theorem
from ..utils import make_safe_filename


def _isinstance_safe(o, t):
    try:
        result = isinstance(o, t)
    except Exception:
        return False
    else:
        return result


Json = Union[dict, list, str, int, float, bool, None]


class _ExtendedEncoder(json.JSONEncoder):
    def default(self, o) -> Json:
        result: Json
        if _isinstance_safe(o, Collection):
            if _isinstance_safe(o, Mapping):
                result = dict(o)
            else:
                result = list(o)
        elif _isinstance_safe(o, datetime):
            result = o.timestamp()
        elif _isinstance_safe(o, UUID):
            result = str(o)
        elif _isinstance_safe(o, Enum):
            result = o.value
        elif _isinstance_safe(o, Decimal):
            result = str(o)
        else:
            result = json.JSONEncoder.default(self, o)
        return result


class MetaData:
    """
    Responsible for interacting with Meta Data files:
    - Accessing attributes such as hashes and timestamps ect.
    - Serialize and deserialize from file format.
    - Generate file hash.
    - Generate meta data hash.
    - TODO: Traverse a dataset's history.
    """

    def __init__(self, file_name: str = '') -> None:
        self.parent_hash = None
        self.file_hash = None
        self.cwd = os.getcwd()

        self.file_name = file_name
        self.timestamp = datetime.now()

        self.experiment_id = Theorem().experiment_id

        if file_name != '':
            try:
                with open(file_name, 'r') as f:
                    self.deserialize(f.read())
            except FileNotFoundError:
                raise FileNotFoundError(f"Metadata not found for {file_name}")
            except Exception as e:
                raise Exception(f"Failed to deserialize '{file_name}: {e}' ")

    def serialize(self) -> str:
        """convert MetaData object to json string."""
        return json.dumps(self.__dict__, cls=_ExtendedEncoder)

    def deserialize(self, raw_json: str):
        """convert a json string to MetaData object."""

        json_obj = json.loads(raw_json)

        for k, v in json_obj.items():
            setattr(self, k, v)

        if isinstance(self.timestamp, datetime):
            self.timestamp = datetime.timestamp(self.timestamp)
        else:
            self.timestamp = datetime.fromtimestamp(self.timestamp)

    @dehydratable
    def save(self, destination: str, cwd: str = '',) -> str:
        """perform a serialization and save to file"""

        if cwd == '':
            cwd = self.cwd

        destination = os.path.join(cwd, destination)

        if not os.path.exists(os.path.dirname(destination)):
            os.makedirs(os.path.dirname(destination))

        logging.debug(f"Saving dataset metadata file to: {destination}")
        with open(destination, "w+") as f:
            logging.debug(f"Saving: {destination}")
            f.write(self.serialize())

        return destination

    def get_dir(self):
        raise NotImplementedError()

    def get_latest(self, subdir_path: str = None):

        newest = None
        now = datetime.min

        metadata_dir = subdir_path or self.get_dir()

        for file in os.listdir(metadata_dir):

            # Ignore keep files.
            if cons.KEEP_FILE in file or os.path.isdir(os.path.join(metadata_dir, file)):
                continue

            full_path = os.path.join(metadata_dir, file)

            # Files can be 1 level nested (See Datasets for an example)
            if os.path.isdir(full_path):
                return self.get_latest(full_path)

            dmd = object.__new__(self.__class__)
            self.__class__.__init__(dmd, full_path)

            if dmd.timestamp > now:
                newest = dmd
                now = dmd.timestamp

        return newest


def gen_meta_file_name_from_hash(meta_data_str, file_name, path):
    meta_data_dir = os.path.join(path, make_safe_filename(file_name))
    meta_hash = gen_meta_hash(meta_data_str)
    meta_data_file_name = meta_hash + ".json"

    return os.path.join(meta_data_dir, meta_data_file_name)


def gen_meta_hash(meta_data_str):
    meta_data_file_name = hashlib.sha1()
    meta_data_file_name.update(str.encode(meta_data_str))
    return meta_data_file_name.hexdigest()



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

from setuptools import setup, find_packages

setup(
    name='aurum',
    version='0.2.0',
    description='Data and Code Versioning for Data Scientists',
    author='Adriano Marques, Nathan Martins, Thales Ribeiro',
    author_email='adriano@xnv.io, nathan@xnv.io, thales@xnv.io',
    python_requires='>=3.7.0',
    include_package_data=True,
    license="GNU LGPLv3",
    install_requires=['nvidia-ml-py3', 'psutil'],
    url='https://github.com/exponential-ventures/aurum',
    packages=find_packages(exclude=['*tests*', 'test*']),
    platforms=['any'],
    entry_points={
        'console_scripts': [
            'au = aurum:main',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

import hashlib
import os
from glob import glob

from .metadata import MetaData, gen_meta_file_name_from_hash
from .. import constants as cons, git
from ..utils import gen_file_hash


class CodeMetaData(MetaData):

    def __init__(self, file_name: str = '') -> None:
        self.file_path_and_hash = None
        super().__init__(file_name)

    def save(self, destination: str = None) -> str:
        destination_path = os.path.join(cons.REPOSITORY_DIR, cons.CODE_METADATA_DIR)
        destination = gen_meta_file_name_from_hash(
            meta_data_str=str(self.timestamp),
            file_name='',
            path=destination_path
        )
        self.file_hash = self.file_path_and_hash
        return super().save(destination)

    def get_dir(self):
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.CODE_METADATA_DIR,
        )


def list_src_files() -> list:
    path = cons.SOURCE_CODE_DIR
    return glob(f"{path}/*")


def generate_src_files_hash_dict() -> dict:
    new_dict = {}
    for p in list_src_files():
        new_dict[p] = gen_file_hash(p)

    return new_dict


def generate_src_files_hash() -> str:
    main_hash = hashlib.sha1()

    for p in list_src_files():
        if os.path.isdir(p):
            continue
        main_hash.update(gen_file_hash(p).encode())

    return main_hash.hexdigest()



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##
import logging
import os

from .metadata import MetaData, gen_meta_file_name_from_hash
from .. import constants as cons, git
from ..utils import gen_dict_hash


class ParameterMetaData(MetaData):

    def __init__(self, file_name: str = '') -> None:
        self.parameters = None
        super().__init__(file_name)

    def save(self, cwd: str, destination: str = None) -> str:
        parent = self.get_latest(
            subdir_path=os.path.join(cwd, cons.REPOSITORY_DIR, cons.PARAMETER_METADATA_DIR)
        )
        self.file_hash = gen_dict_hash(self.parameters)

        if parent and self.file_hash != parent.file_hash:
            self.parent_hash = parent.file_hash

        destination = gen_meta_file_name_from_hash(
            meta_data_str=str(self.timestamp),
            file_name='',
            path=self.get_dir(),
        )

        logging.debug(f"Saving parameters file to: {destination}")
        return super().save(destination=destination, cwd=cwd)

    @property
    def parameter_hash(self) -> str:
        return gen_dict_hash(self.parameters)

    def get_dir(self):
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.PARAMETER_METADATA_DIR,
        )



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##

__author__ = "Adriano Marques, Nathan Martins, Thales Ribeiro"
__copyright__ = "Copyright (C) 2019 Exponential Ventures LLC"
__license__ = "GNU LESSER GENERAL PUBLIC LICENSE 2.0"
__url__ = "https://github.com/exponential-ventures/aurum"
__version__ = "0.1"

import argparse

from aurum.base import execute_commands
import aurum.constants as cons


def config_parser() -> argparse.ArgumentParser:
    description = """Aurum is a new and simplified approach for data scientists to
    keep track of data and code without having to get another PhD for it. Aurum
    keeps track of all code and data changes, and lets you easily reproduce any
    experiment as well as easily compare metrics across experiments.
    """
    epilog = "And that's how you make your live easier. You're welcome."
    parser = argparse.ArgumentParser(description=description, epilog=epilog, add_help=True)

    subparsers = parser.add_subparsers()

    #######
    # init
    parser_init = subparsers.add_parser(cons.INIT, help="initialize current directory as an aurum repository")
    parser_init.set_defaults(subcommand=cons.INIT)

    #######
    # load
    parser_load = subparsers.add_parser("load", help="load a specific experiment by tag")
    parser_load.set_defaults(subcommand="load")
    parser_load.add_argument("tag", type=str, action="store")

    #######
    # data
    parser_data = subparsers.add_parser(cons.DATA, help="operations on data files")
    parser_data.set_defaults(subcommand=cons.DATA)

    subparser_data = parser_data.add_subparsers()

    parser_data_add = subparser_data.add_parser(cons.DATA_ADD, help="add specified data file(s) to the data index")
    parser_data_add.set_defaults(subcommand2=cons.DATA_ADD)
    parser_data_add.add_argument("files", type=str, nargs='+')

    ######
    # add
    parser_add = subparsers.add_parser(cons.DATA_ADD, help="add specified file to the index")
    parser_add.set_defaults(subcommand=cons.DATA_ADD)

    subparser_add = parser_add.add_subparsers()

    parser_add_files = subparser_add.add_parser(cons.DATA_ADD, help="add specified file(s) to the data index")
    parser_add_files.set_defaults(subcommand2=cons.DATA_ADD)
    parser_add_files.add_argument("files", type=str, nargs="+")

    #######
    # data rm
    parser_data_rm = subparser_data.add_parser(cons.DATA_RM, help="remove specified data file(s) from the data index")
    parser_data_rm.set_defaults(subcommand2=cons.DATA_RM)
    parser_data_rm.add_argument("files", type=str, nargs="+")

    deletion_type_parser = parser_data_rm.add_mutually_exclusive_group(required=False)
    deletion_type_parser.add_argument("--soft-delete", dest="soft_delete", action="store_true")
    deletion_type_parser.add_argument("--hard-delete", dest="soft_delete", action="store_false")
    parser_data_rm.set_defaults(soft_delete=True)

    #######
    # metrics
    parser_metrics = subparsers.add_parser(cons.METRICS, help="display metrics")
    parser_metrics.set_defaults(subcommand=cons.METRICS)
    parser_metrics.add_argument(cons.EXPERIMENT_IDS, type=str, nargs='?', default=None)

    #######
    # export tag
    parser_export = subparsers.add_parser(cons.EXPORT_TAG, help="export tag")
    parser_export.set_defaults(subcommand=cons.EXPORT_TAG)
    parser_export.add_argument("tag", type=str, action="store")
    parser_export.add_argument("--no-data", dest="no_data", action="store_true", default=False)
    parser_export.add_argument("--no-metrics", dest="no_metrics", action="store_true", default=False)
    parser_export.add_argument("--no-logs", dest="no_logs", action="store_true", default=False)


    ######
    # Other arguments
    parser.add_argument("--verbose", help="increase output verbosity", action="store_true")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    return parser


def main() -> None:
    execute_commands(config_parser())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


from .metadata import MetaData, gen_meta_hash, gen_meta_file_name_from_hash
from .code import CodeMetaData, generate_src_files_hash_dict, generate_src_files_hash
from .dataset import DatasetMetaData
from .experiment import ExperimentMetaData
from .metrics import MetricsMetaData
from .weights import WeightsMetaData
from .parameters import ParameterMetaData
from .requirements import RequirementsMetaData



#!/usr/bin/env python3
##
## Authors: Rhuan Karlus
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import logging
import os
import time

from .metadata import MetaData, gen_meta_file_name_from_hash
from .. import constants as cons
from ..theorem import Theorem
from ..utils import gen_dict_hash


class WeightsMetaData(MetaData):

    def __init__(self, file_name: str = '') -> None:
        self.weights = None
        self.binary_file_path = ''
        self.relative_path = ''
        super().__init__(file_name)

    def save(self, cwd: str, destination: str = None) -> str:
        current_timestamp = str(round(time.time() * 1000))

        if Theorem().has_any_change():
            self.file_hash = gen_dict_hash({
                'binary_file_path': self.binary_file_path,
            })
            parent = self.get_latest(subdir_path=os.path.join(cwd, cons.REPOSITORY_DIR, cons.WEIGHTS_METADATA_DIR))
            if parent and self.file_hash != parent.file_hash:
                self.parent_hash = parent.file_hash

        if destination is None:
            destination = gen_meta_file_name_from_hash(str(current_timestamp), '', self.get_dir())

        logging.debug(f"Saving model file to: {destination}")
        return super().save(destination=destination, cwd=cwd)

    def save_binary(self, encoded_model: bytes, cwd: str = '') -> str:

        current_timestamp = str(round(time.time() * 1000))

        if cwd == '':
            cwd = os.getcwd()

        self.relative_path = os.path.join(self.get_binaries_dir(), current_timestamp)
        self.binary_file_path = os.path.join(cwd, self.relative_path)

        with open(self.binary_file_path, 'wb') as f:
            f.write(encoded_model)

        return self.binary_file_path

    def get_dir(self):
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.WEIGHTS_METADATA_DIR,
        )

    def load_binary(self):

        if self.relative_path != '':
            try:
                with open(self.relative_path, mode='rb') as f:
                    return f.read()
            except FileNotFoundError:
                logging.warning(f"tried loading from relative path: '{self.relative_path}' but failed")
                with open(self.binary_file_path, mode='rb') as f:
                    return f.read()

    @staticmethod
    def get_binaries_dir():
        return os.path.join(
            cons.REPOSITORY_DIR,
            cons.WEIGHTS_METADATA_DIR,
            cons.WEIGHTS_BINARIES_DIR,
        )



#!/usr/bin/env python3
##
## Authors: Adriano Marques
##          Nathan Martins
##          Thales Ribeiro
##
## Copyright (C) 2019 Exponential Ventures LLC
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free Software
##    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
##


import sys
import logging
from tempfile import TemporaryFile

from .singleton import SingletonDecorator


class tee:
    """
    Inspired by: http://shallowsky.com/blog/programming/python-tee.html
    """

    def __init__(self, stream):
        self.temp_file = TemporaryFile(mode="w+")
        self.stream = stream

    def write(self, text):
        self.temp_file.write(text)
        self.stream.write(text)

    def flush(self):
        self.temp_file.flush()
        self.stream.flush()


@SingletonDecorator
class LoggingTracker:

    def __init__(self) -> None:
        super().__init__()

        self.stdout = tee(sys.stdout)
        self.stderr = tee(sys.stderr)
        logging.getLogger().addHandler(logging.StreamHandler(self.stdout.temp_file))

        sys.stdout = self.stdout
        sys.stderr = self.stderr
